import logging
import os
from io import StringIO
from typing import TextIO

import regex
from ruamel.yaml import YAML

from zabbixci.settings import Settings

yaml = YAML()

logger = logging.getLogger(__name__)


class Template:
    """
    A Python representation of a Zabbix template
    """

    _export: dict
    _level: int = None

    @property
    def is_template(self):
        return "templates" in self._export

    @property
    def _template(self):
        return self._export["templates"][0]

    @property
    def groups(self):
        return map(lambda group: group["name"], self._template["groups"])

    @property
    def name(self):
        return self._template["name"]

    @property
    def uuid(self):
        return self._template["uuid"]

    @property
    def template_id(self):
        return self._template["templateid"]

    @property
    def primary_group(self):
        """
        The most specific group of the template, the lowest child in the hierarchy
        """
        selected_group = self._template["groups"][0]["name"]
        selected_length = 0

        # Most specific group is based on the group with the most slashes,
        # specifying the lowest child in the hierarchy
        for group in self._template["groups"]:
            name: str = group["name"]
            if not Settings.ROOT_TEMPLATE_GROUP in name:
                continue

            split = regex.split(r"\/+", name)

            if len(split) > selected_length:
                selected_group = name
                selected_length = len(split)

        return selected_group

    @property
    def truncated_groups(self):
        """
        The primary group of the template, without the root group
        """
        # split = regex.split(fr'{PARENT_GROUP}\/+', self.primary_group)
        match_group = regex.match(
            rf"{Settings.ROOT_TEMPLATE_GROUP}\/+(.+)", self.primary_group
        )

        return match_group.group(1) if match_group else ""

    @property
    def linked_templates(self):
        return (
            [t["name"] for t in self._template["templates"]]
            if "templates" in self._template
            else []
        )

    @property
    def zabbix_version(self):
        return self._export["version"]

    def __init__(self, export: dict):
        self._export = export

    def __str__(self):
        return self._template["name"]

    def _yaml_dump(self, stream: TextIO):
        """
        Dump Zabbix importable template to stream
        """
        yaml.dump({"zabbix_export": self._export}, stream)

    def save(self):
        """
        Save the template to the cache
        """
        os.makedirs(
            f"{Settings.CACHE_PATH}/{Settings.TEMPLATE_PREFIX_PATH}/{self.truncated_groups}",
            exist_ok=True,
        )

        with open(
            f"{Settings.CACHE_PATH}/{Settings.TEMPLATE_PREFIX_PATH}/{self.truncated_groups}/{self._template['name']}.yaml",
            "w",
        ) as file:
            self._yaml_dump(file)

    def export(self):
        """
        Export the template as Zabbix importable YAML
        """
        stream = StringIO()
        self._yaml_dump(stream)
        return stream.getvalue()

    def level(self, templates: list):
        """
        Get the amount of linked levels
        """
        logger.debug(f"Getting level for {self.name}")

        linked_templates = [t for t in templates if t.name in self.linked_templates]

        self._level = (
            max([template.level(templates) for template in linked_templates] or [0]) + 1
        )

        return self._level

    @staticmethod
    def open(path: str):
        """
        Open a template from the cache
        """
        with open(f"{Settings.CACHE_PATH}/{path}", "r") as file:
            return Template(yaml.load(file)["zabbix_export"])

    @staticmethod
    def from_zabbix(template: dict, template_groups: dict, zabbix_version: str):
        """
        Create a individual template from a bulk Zabbix export
        """
        groups = list(
            filter(
                lambda group: group["name"]
                in [group["name"] for group in template["groups"]],
                template_groups,
            )
        )

        return Template(
            {
                "version": zabbix_version,
                "template_groups": groups,
                "templates": [template],
            }
        )



import os

from ruamel.yaml import YAML

yaml = YAML()


class Settings:
    ZABBIX_URL = "http://localhost:8080"
    ZABBIX_USER = None
    ZABBIX_PASSWORD = None
    ZABBIX_TOKEN = None
    REMOTE = None
    ROOT_TEMPLATE_GROUP = "Templates"
    GIT_AUTHOR_NAME = "Zabbix CI"
    GIT_AUTHOR_EMAIL = "zabbixci@localhost"
    PULL_BRANCH = "main"
    PUSH_BRANCH = "main"
    TEMPLATE_PREFIX_PATH = ""
    TEMPLATE_WHITELIST = []
    TEMPLATE_BLACKLIST = []
    CACHE_PATH = "./cache"
    BATCH_SIZE = 50
    IGNORE_TEMPLATE_VERSION = False
    INSECURE_SSL_VERIFY = False
    GIT_USERNAME = "git"
    GIT_PASSWORD = None
    GIT_PUBKEY = None
    GIT_PRIVKEY = None
    GIT_KEYPASSPHRASE = None
    DRY_RUN = False

    @classmethod
    def from_env(cls):
        for key in cls.__dict__.keys():
            if key in os.environ:
                setattr(cls, key, os.environ[key])

    @classmethod
    def read_config(cls, path):
        with open(path, "r") as f:
            data = yaml.load(f)
            for key, value in data.items():
                setattr(cls, key.upper(), value)



import logging
import os
import timeit
from io import StringIO

import pygit2
from pygit2.enums import FileStatus, ResetMode
from regex import search
from ruamel.yaml import YAML

from zabbixci.settings import Settings
from zabbixci.utils.git import Git
from zabbixci.utils.template import Template
from zabbixci.utils.zabbix import Zabbix


class ZabbixCI:
    logger = logging.getLogger(__name__)
    yaml = YAML()

    _settings = None
    _git_cb = None

    _zabbix = None
    _git = None

    def __init__(self, settings=Settings, logger=None):
        self._settings = settings

        if logger:
            self.logger = logger

        self.create_git_callback()
        self.create_zabbix()
        self.create_git()

    def create_git_callback(self):
        """
        Create a pygit2 RemoteCallbacks object with the appropriate credentials
        Handles both username/password and SSH keypair authentication

        :param settings: Settings object
        """
        if self._settings.GIT_USERNAME and self._settings.GIT_PASSWORD:
            self.logger.info("Using username and password for authentication")
            credentials = pygit2.UserPass(
                self._settings.GIT_USERNAME, self._settings.GIT_PASSWORD
            )
        elif self._settings.GIT_PUBKEY and self._settings.GIT_PRIVKEY:
            self.logger.info("Using SSH keypair for authentication")
            credentials = pygit2.Keypair(
                self._settings.GIT_USERNAME,
                self._settings.GIT_PUBKEY,
                self._settings.GIT_PRIVKEY,
                self._settings.GIT_KEYPASSPHRASE,
            )
        else:
            self.logger.info("Using SSH agent for authentication")
            credentials = pygit2.KeypairFromAgent(self._settings.GIT_USERNAME)

        self._git_cb = pygit2.RemoteCallbacks(credentials=credentials)

        if self._settings.INSECURE_SSL_VERIFY:
            self._git_cb.certificate_check = lambda *args, **kwargs: True

    def create_zabbix(self):
        """
        Create a Zabbix object with the appropriate credentials
        """
        self._zabbix = Zabbix(
            url=self._settings.ZABBIX_URL,
            user=self._settings.ZABBIX_USER,
            password=self._settings.ZABBIX_PASSWORD,
            token=self._settings.ZABBIX_TOKEN,
            validate_certs=not self._settings.INSECURE_SSL_VERIFY,
        )

    def create_git(self):
        """
        Create a Git object with the appropriate credentials
        """
        self._git = Git(self._settings.CACHE_PATH, self._git_cb)

    def push(self):
        """
        Fetch Zabbix state and commit changes to git remote
        """
        self._git.fetch(Settings.REMOTE, self._git_cb)

        if not self._git.is_empty:
            # If the repository is empty, new branches can't be created. But it is
            # safe to push to the default branch
            self._git.switch_branch(Settings.PUSH_BRANCH)

            # Pull the latest remote state
            try:
                self._git.pull(Settings.REMOTE, self._git_cb)
            except KeyError:
                self.logger.info(
                    f"Remote branch does not exist, using state from branch {Settings.PULL_BRANCH}"
                )
                # Remote branch does not exist, we pull the default branch and create a new branch
                self._git.switch_branch(Settings.PULL_BRANCH)
                self._git.pull(Settings.REMOTE, self._git_cb)

                # Create a new branch
                self._git.switch_branch(Settings.PUSH_BRANCH)

        # Reflect current Zabbix state in the cache
        self.cleanup_cache()
        self.zabbix_to_file()

        # Check if there are any changes to commit
        if not self._git.has_changes and not self._git.ahead_of_remote:
            self.logger.info("No changes detected")
            exit()

        self.logger.info("Remote differs from local state, preparing to push")

        # Commit and push the changes
        self._git.add_all()

        host = os.getenv(
            "ZABBIX_HOST", search("https?://([^/]+)", self._zabbix.zapi.url).group(1)
        )

        if self._git.has_changes:
            # Create a commit
            changes = self._git.diff()
            files = [patch.delta.new_file.path for patch in changes]

            for file in files:
                self.logger.info(f"Detected change in {file}")

            # Generate commit message
            self._git.commit(f"Committed Zabbix state from {host}")
            self.logger.info(
                f"Staged changes from {host} committed to {self._git.current_branch}"
            )
        else:
            self.logger.info("No staged changes, updating remote with current state")

        if not self._settings.DRY_RUN:
            self._git.push(Settings.REMOTE, self._git_cb)

    def pull(self):
        """
        Pull current state from git remote and update Zabbix
        """
        self._git.switch_branch(Settings.PULL_BRANCH)

        # Pull the latest remote state, untracked changes are preserved
        self._git.pull(Settings.REMOTE, self._git_cb)
        self._git.reset(
            self._git._repository.lookup_reference(
                f"refs/remotes/origin/{Settings.PULL_BRANCH}"
            ).target,
            ResetMode.HARD,
        )

        current_revision = self._git.get_current_revision()

        # Reflect current Zabbix state in the cache
        self.cleanup_cache()
        self.zabbix_to_file()

        zabbix_version = self._zabbix.get_server_version()

        # Check for untracked changes, if there are any, we know Zabbix is out of
        # sync
        if self._git.has_changes:
            self.logger.info(
                "Detected local file changes, detecting changes for zabbix sync"
            )

        status = self._git.status()

        # Get the changed files, we compare the untracked changes with the desired.
        # When we have a new untracked file, that means it was deleted in the desired state.
        files = [
            path
            for path, flags in status.items()
            if flags in [FileStatus.WT_DELETED, FileStatus.WT_MODIFIED]
        ]
        deleted_files = [
            path for path, flags in status.items() if flags == FileStatus.WT_NEW
        ]

        self.logger.debug(f"Following templates have changed on Git: {files}")
        self.logger.debug(f"Following templates are deleted from Git {deleted_files}")

        # Sync the file cache with the desired git state
        self._git.reset(current_revision, ResetMode.HARD)

        templates: list[Template] = []

        # Open the changed files
        for file in files:
            if not file.endswith(".yaml"):
                continue

            # Check if file is within the desired path
            if not file.startswith(Settings.TEMPLATE_PREFIX_PATH):
                self.logger.debug(f"Skipping .yaml file {file} outside of prefix path")
                continue

            template = Template.open(file)

            if not template or not template.is_template:
                self.logger.warning(f"Could load file {file} as a template")
                continue

            if self.ignore_template(template.name):
                continue

            if (
                not Settings.IGNORE_TEMPLATE_VERSION
                and template.zabbix_version.split(".")[0:2]
                != zabbix_version.split(".")[0:2]
            ):
                self.logger.warning(
                    f"Template {template.name}: {template.zabbix_version} must match major Zabbix version {'.'.join(zabbix_version.split('.')[0:2])}"
                )
                continue

            templates.append(template)
            self.logger.info(f"Detected change in {template.name}")

        if len(templates):
            tic = timeit.default_timer()

            # Group templates by level
            templates = sorted(
                templates, key=lambda template: template.level(templates)
            )

            toc = timeit.default_timer()
            self.logger.info("Sorted templates in {:.2f}s".format(toc - tic))

            # Import the templates
            for template in templates:
                self.logger.info(f"Importing {template.name}, level {template._level}")

                if not self._settings.DRY_RUN:
                    self._zabbix.import_template(template)

        template_names = []

        # Delete the deleted templates
        for file in deleted_files:
            template = Template.open(file)

            if not template or not template.is_template:
                self.logger.warning(f"Could not open to be deleted file {file}")
                continue

            if self.ignore_template(template.name):
                continue

            if template.uuid in [t.uuid for t in templates]:
                self.logger.debug(
                    f"Template {template.name} is being imported under a different name or path, skipping deletion"
                )
                continue

            template_names.append(template.name)
            self.logger.info(f"Added {template.name} to deletion queue")

        if len(template_names):
            self.logger.info(f"Deleting {len(template_names)} templates from Zabbix")
            template_ids = [
                t["templateid"] for t in self._zabbix.get_templates_name(template_names)
            ]

            if len(template_ids):
                if not self._settings.DRY_RUN:
                    self._zabbix.delete_template(template_ids)

        # clean local changes
        self._git.clean()

    def zabbix_to_file(self) -> None:
        """
        Export Zabbix templates to the cache
        """
        templates = self._zabbix.get_templates([Settings.ROOT_TEMPLATE_GROUP])

        self.logger.info(f"Found {len(templates)} templates in Zabbix")
        self.logger.debug(f"Found Zabbix templates: {[t['name'] for t in templates]}")

        # Split by Settings.BATCH_SIZE
        batches = [
            templates[i : i + Settings.BATCH_SIZE]
            for i in range(0, len(templates), Settings.BATCH_SIZE)
        ]

        for index, batch in enumerate(batches):
            self.logger.info(
                f"Processing export batch {index + 1}/{len(batches)} [{(index * Settings.BATCH_SIZE) + 1}/{len(templates)}]"
            )

            # Get the templates
            template_yaml = self._zabbix.export_template(
                [template["templateid"] for template in batch]
            )

            export_yaml = self.yaml.load(StringIO(template_yaml))

            if not "templates" in export_yaml["zabbix_export"]:
                self.logger.info("No templates found in Zabbix")
                return

            # Write the templates to the cache
            for template in export_yaml["zabbix_export"]["templates"]:
                template = Template.from_zabbix(
                    template,
                    export_yaml["zabbix_export"]["template_groups"],
                    export_yaml["zabbix_export"]["version"],
                )

                if self.ignore_template(template.name):
                    continue

                template.save()

    @classmethod
    def cleanup_cache(cls, full: bool = False) -> None:
        """
        Clean all .yaml (template) files from the cache directory

        If full is True, also remove the .git directory and all other files
        """
        for root, dirs, files in os.walk(
            f"{Settings.CACHE_PATH}/{Settings.TEMPLATE_PREFIX_PATH}", topdown=False
        ):
            if f"{Settings.CACHE_PATH}/.git" in root and not full:
                continue

            for name in files:
                if name.endswith(".yaml") or full:
                    os.remove(os.path.join(root, name))

            for name in dirs:
                if name == ".git" and root == Settings.CACHE_PATH and not full:
                    continue

                # Remove empty directories
                if not os.listdir(os.path.join(root, name)):
                    os.rmdir(os.path.join(root, name))

        if full:
            os.rmdir(Settings.CACHE_PATH)
            cls.logger.info("Cache directory cleared")

    @classmethod
    def ignore_template(cls, template_name: str) -> bool:
        """
        Returns true if template should be ignored because of the blacklist or whitelist
        """
        if template_name in Settings.TEMPLATE_BLACKLIST:
            cls.logger.debug(f"Skipping blacklisted template {template_name}")
            return True

        if (
            len(Settings.TEMPLATE_WHITELIST)
            and template_name not in Settings.TEMPLATE_WHITELIST
        ):
            cls.logger.debug(f"Skipping non whitelisted template {template_name}")
            return True

        return False



from typing import ParamSpec

from ruamel.yaml import YAML
from zabbix_utils import ZabbixAPI

from zabbixci.utils.template import Template

yaml = YAML()

P = ParamSpec("P")


class Zabbix:
    zapi = None

    def __init__(self, *args: P.args, **kwargs: P.kwargs):
        self.zapi = ZabbixAPI(*args, **kwargs)

    def _get_template_group(self, template_group_names: list[str]):
        return self.zapi.send_api_request(
            "templategroup.get", {"search": {"name": template_group_names}}
        )["result"]

    def get_templates(self, template_group_names: list[str]):
        ids = self._get_template_group(template_group_names)

        template_group_ids = [group["groupid"] for group in ids]

        return self.zapi.send_api_request(
            "template.get", {"groupids": template_group_ids}
        )["result"]

    def export_template(self, template_ids: list[int]):
        return self.zapi.send_api_request(
            "configuration.export",
            {"options": {"templates": template_ids}, "format": "yaml"},
        )["result"]

    def import_template(self, template: Template):
        export = template.export()

        return self.zapi.send_api_request(
            "configuration.import",
            {
                "format": "yaml",
                "rules": {
                    "template_groups": {"createMissing": True, "updateExisting": True},
                    "templateLinkage": {"createMissing": True, "deleteMissing": True},
                    "templates": {
                        "createMissing": True,
                        "updateExisting": True,
                    },
                },
                "source": export,
            },
        )["result"]

    def get_server_version(self):
        return self.zapi.send_api_request("apiinfo.version", need_auth=False)["result"]

    def get_templates_name(self, name: list[str]):
        return self.zapi.send_api_request("template.get", {"filter": {"host": name}})[
            "result"
        ]

    def delete_template(self, template_ids: list[int]):
        return self.zapi.send_api_request("template.delete", template_ids)["result"]



import argparse
import logging
import logging.config

from zabbixci.settings import Settings
from zabbixci.zabbixci import ZabbixCI

# Read command line arguments to fill the settings

logger = logging.getLogger(__name__)


def read_args():
    parser = argparse.ArgumentParser(
        description="""ZabbixCI is a tool to manage Zabbix templates in a Git repository.
        
        ZabbixCI adds version control to Zabbix templates, allowing you to track changes, synchronize templates between different Zabbix servers, and collaborate with other team members.""",
        prog="zabbixci",
    )
    parser.add_argument(
        "action",
        help="The action to perform",
        choices=["push", "pull", "clearcache"],
    )

    # Provide configuration as file
    parser.add_argument(
        "--config",
        help="Provide configuration as yaml file",
    )

    # Zabbix
    parser.add_argument(
        "--zabbix-url",
        help="Zabbix URL",
    )
    parser.add_argument(
        "--zabbix-user",
        help="Zabbix user for user/password authentication",
    )
    parser.add_argument(
        "--zabbix-password",
        help="Zabbix password for user/password authentication",
    )
    parser.add_argument(
        "--zabbix-token",
        help="Zabbix token for token authentication (preferred)",
    )

    # Git
    parser.add_argument(
        "--remote",
        help="URL of the remote git repository, supports ssh and http(s)",
    )
    parser.add_argument(
        "--pull-branch",
        help="Branch to pull from",
    )
    parser.add_argument(
        "--push-branch",
        help="Branch to push to",
    )
    parser.add_argument(
        "--git-username",
        help="Git username, used for http(s) authentication",
    )
    parser.add_argument(
        "--git-password",
        help="Git password, used for http(s) authentication",
    )
    parser.add_argument(
        "--git-pubkey",
        help="SSH public key, used for ssh authentication",
    )
    parser.add_argument(
        "--git-privkey",
        help="SSH private key, used for ssh authentication",
    )
    parser.add_argument(
        "--git-keypassphrase",
        help="SSH key passphrase, used for ssh authentication",
    )

    # ZabbixCI
    parser.add_argument(
        "--root-template-group",
        help="Zabbix Template Group root, defaults to Templates",
    )
    parser.add_argument(
        "--template-prefix-path",
        help="The path in the git repository, used to store the templates",
    )
    parser.add_argument(
        "--template-whitelist",
        help="Comma separated list of templates to include",
    )
    parser.add_argument(
        "--template-blacklist",
        help="Comma separated list of templates to exclude",
    )
    parser.add_argument(
        "--cache",
        help="Cache path for git repository, defaults to ./cache",
    )
    parser.add_argument(
        "--dry-run",
        help="Dry run, only show changes",
        action="store_true",
        default=None,
    )

    # ZabbixCI advanced
    parser.add_argument(
        "-v",
        help="Enable verbose logging",
        action="store_true",
        dest="verbose",
    )
    parser.add_argument(
        "-vv",
        "--debug",
        help="Enable debug logging",
        action="store_true",
        dest="debug",
    )
    parser.add_argument(
        "-vvv",
        "--debug-all",
        help="Enable debug logging for all modules",
        action="store_true",
        dest="debug_all",
    )
    parser.add_argument(
        "--batch-size",
        help="Batch size for Zabbix API export requests",
    )
    parser.add_argument(
        "--ignore-template-version",
        help="Ignore template versions on import, useful for initial import",
        action="store_true",
        default=None,
    )
    parser.add_argument(
        "--insecure-ssl-verify",
        help="Disable SSL verification for Zabbix API and git, only use for testing",
        action="store_true",
        default=None,
    )

    return parser.parse_args()


def parse_cli():
    Settings.from_env()
    args = read_args()
    arguments = vars(args)

    global_level = (
        logging.DEBUG
        if args.debug_all
        else logging.INFO if args.verbose else logging.WARN
    )

    logging.basicConfig(
        format="%(asctime)s [%(name)s]  [%(levelname)s]: %(message)s",
        level=global_level,
    )

    zabbixci_logger = logging.getLogger("zabbixci")
    zabbixci_logger.setLevel(
        logging.DEBUG
        if args.debug or args.debug_all
        else logging.INFO if args.verbose else logging.WARN
    )

    if args.config:
        Settings.read_config(args.config)

    for key, value in arguments.items():
        if value is not None:
            logger.debug(f"Setting {key} to {value}")
            setattr(Settings, key.upper(), value)

    settings_debug = {
        **Settings.__dict__,
        "ZABBIX_PASSWORD": "********",
        "ZABBIX_TOKEN": "********",
        "REMOTE": "********",
    }

    logger.debug(f"Settings: {settings_debug}")

    if args.action == "push":
        zabbixci = ZabbixCI()
        zabbixci.push()

    elif args.action == "pull":
        zabbixci = ZabbixCI()
        zabbixci.pull()

    elif args.action == "clearcache":
        ZabbixCI.cleanup_cache(full=True)


if __name__ == "__main__":
    parse_cli()
